public inbox for linux-ext4@vger.kernel.org
 help / color / mirror / Atom feed
From: Andrey Albershteyn <aalbersh@kernel.org>
To: linux-xfs@vger.kernel.org, fsverity@lists.linux.dev,
	linux-fsdevel@vger.kernel.org, ebiggers@kernel.org
Cc: Andrey Albershteyn <aalbersh@kernel.org>,
	hch@lst.de, linux-ext4@vger.kernel.org,
	linux-f2fs-devel@lists.sourceforge.net,
	linux-btrfs@vger.kernel.org, djwong@kernel.org
Subject: [PATCH v6 06/22] iomap: introduce IOMAP_F_FSVERITY and teach writeback to handle fsverity
Date: Tue, 31 Mar 2026 23:28:07 +0200	[thread overview]
Message-ID: <20260331212827.2631020-7-aalbersh@kernel.org> (raw)
In-Reply-To: <20260331212827.2631020-1-aalbersh@kernel.org>

This flag indicates that I/O is for fsverity metadata.

In the write path skip i_size check and i_size updates as metadata is
past EOF. In writeback don't update i_size and continue writeback if
even folio is beyond EOF. In read path don't zero fsverity folios, again
they are past EOF.

The iomap_block_needs_zeroing() is also called from write path. For
folios of larger order we don't want to zero out pages in the folio as
these could contain other merkle tree blocks. For fsverity, filesystem
will request to read PAGE_SIZE memory regions. For data folios, iomap
will zero the rest of the folio for anything which is beyond EOF. We
don't want this for fsverity folios.

Signed-off-by: Andrey Albershteyn <aalbersh@kernel.org>
Reviewed-by: "Darrick J. Wong" <djwong@kernel.org>
---
 fs/iomap/buffered-io.c | 43 +++++++++++++++++++++++++++++++++---------
 fs/iomap/trace.h       |  3 ++-
 include/linux/iomap.h  |  8 ++++++++
 3 files changed, 44 insertions(+), 10 deletions(-)

diff --git a/fs/iomap/trace.h b/fs/iomap/trace.h
index 532787277b16..5252051cc137 100644
--- a/fs/iomap/trace.h
+++ b/fs/iomap/trace.h
@@ -118,7 +118,8 @@ DEFINE_RANGE_EVENT(iomap_zero_iter);
 	{ IOMAP_F_ATOMIC_BIO,	"ATOMIC_BIO" }, \
 	{ IOMAP_F_PRIVATE,	"PRIVATE" }, \
 	{ IOMAP_F_SIZE_CHANGED,	"SIZE_CHANGED" }, \
-	{ IOMAP_F_STALE,	"STALE" }
+	{ IOMAP_F_STALE,	"STALE" }, \
+	{ IOMAP_F_FSVERITY,	"FSVERITY" }
 
 
 #define IOMAP_DIO_STRINGS \
diff --git a/include/linux/iomap.h b/include/linux/iomap.h
index 531f9ebdeeae..4506a99d5285 100644
--- a/include/linux/iomap.h
+++ b/include/linux/iomap.h
@@ -87,6 +87,14 @@ struct vm_fault;
 #define IOMAP_F_INTEGRITY	0
 #endif /* CONFIG_BLK_DEV_INTEGRITY */
 
+/*
+ * Indicates reads and writes of fsverity metadata.
+ *
+ * Fsverity metadata is stored after the regular file data and thus beyond
+ * i_size.
+ */
+#define IOMAP_F_FSVERITY	(1U << 10)
+
 /*
  * Flag reserved for file system specific usage
  */
diff --git a/fs/iomap/buffered-io.c b/fs/iomap/buffered-io.c
index e4b6886e5c3c..a80fcb598cc8 100644
--- a/fs/iomap/buffered-io.c
+++ b/fs/iomap/buffered-io.c
@@ -353,9 +353,26 @@ static inline bool iomap_block_needs_zeroing(const struct iomap_iter *iter,
 {
 	const struct iomap *srcmap = iomap_iter_srcmap(iter);
 
-	return srcmap->type != IOMAP_MAPPED ||
-		(srcmap->flags & IOMAP_F_NEW) ||
-		pos >= i_size_read(iter->inode);
+	/*
+	 * If this block has not been written, there's nothing to read
+	 */
+	if (srcmap->type != IOMAP_MAPPED)
+		return true;
+
+	/*
+	 * Newly allocated blocks have not been written
+	 */
+	if (srcmap->flags & IOMAP_F_NEW)
+		return true;
+
+	/*
+	 * fsverity metadata is stored past i_size, we need to read it instead of
+	 * zeroing
+	 */
+	if (srcmap->flags & IOMAP_F_FSVERITY)
+		return false;
+
+	return pos >= i_size_read(iter->inode);
 }
 
 /**
@@ -1167,13 +1184,14 @@ static int iomap_write_iter(struct iomap_iter *iter, struct iov_iter *i,
 		 * unlock and release the folio.
 		 */
 		old_size = iter->inode->i_size;
-		if (pos + written > old_size) {
+		if (pos + written > old_size &&
+		    !(iter->iomap.flags & IOMAP_F_FSVERITY)) {
 			i_size_write(iter->inode, pos + written);
 			iter->iomap.flags |= IOMAP_F_SIZE_CHANGED;
 		}
 		__iomap_put_folio(iter, write_ops, written, folio);
 
-		if (old_size < pos)
+		if (old_size < pos && !(iter->iomap.flags & IOMAP_F_FSVERITY))
 			pagecache_isize_extended(iter->inode, old_size, pos);
 
 		cond_resched();
@@ -1801,13 +1819,20 @@ static int iomap_writeback_range(struct iomap_writepage_ctx *wpc,
  * Check interaction of the folio with the file end.
  *
  * If the folio is entirely beyond i_size, return false.  If it straddles
  * i_size, adjust end_pos and zero all data beyond i_size.
  */
-static bool iomap_writeback_handle_eof(struct folio *folio, struct inode *inode,
-		u64 *end_pos)
+static bool iomap_writeback_handle_eof(struct folio *folio,
+		struct iomap_writepage_ctx *wpc, u64 *end_pos)
 {
+	struct inode *inode = wpc->inode;
 	u64 isize = i_size_read(inode);
 
+	if (wpc->iomap.flags & IOMAP_F_FSVERITY) {
+		WARN_ON_ONCE(folio_pos(folio) < isize);
+		return true;
+	}
+
 	if (*end_pos > isize) {
 		size_t poff = offset_in_folio(folio, isize);
 		pgoff_t end_index = isize >> PAGE_SHIFT;
@@ -1873,7 +1898,7 @@ int iomap_writeback_folio(struct iomap_writepage_ctx *wpc, struct folio *folio)
 
 	trace_iomap_writeback_folio(inode, pos, folio_size(folio));
 
-	if (!iomap_writeback_handle_eof(folio, inode, &end_pos))
+	if (!iomap_writeback_handle_eof(folio, wpc, &end_pos))
 		return 0;
 	WARN_ON_ONCE(end_pos <= pos);
 
@@ -1801,13 +1819,20 @@ static int iomap_writeback_range(struct iomap_writepage_ctx *wpc,
  *
  * If the folio is entirely beyond i_size, return false.  If it straddles
- * i_size, adjust end_pos and zero all data beyond i_size.
+ * i_size, adjust end_pos and zero all data beyond i_size. Don't skip fsverity
+ * folios as those are beyond i_size.
  */
-- 
2.51.2


  parent reply	other threads:[~2026-03-31 21:28 UTC|newest]

Thread overview: 43+ messages / expand[flat|nested]  mbox.gz  Atom feed  top
2026-03-31 21:28 [PATCH v6 00/22] fs-verity support for XFS with post EOF merkle tree Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 01/22] fsverity: report validation errors through fserror to fsnotify Andrey Albershteyn
2026-04-01 21:19   ` Eric Biggers
2026-03-31 21:28 ` [PATCH v6 02/22] fsverity: expose ensure_fsverity_info() Andrey Albershteyn
2026-04-01  6:27   ` Christoph Hellwig
2026-04-01 22:02   ` Eric Biggers
2026-04-02 14:02     ` Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 03/22] fsverity: generate and store zero-block hash Andrey Albershteyn
2026-04-01 22:27   ` Eric Biggers
2026-04-02 14:47     ` Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 04/22] fsverity: pass digest size and hash of the empty block to ->write Andrey Albershteyn
2026-04-01 23:36   ` Eric Biggers
2026-03-31 21:28 ` [PATCH v6 05/22] fsverity: hoist pagecache_read from f2fs/ext4 to fsverity Andrey Albershteyn
2026-04-01 23:44   ` Eric Biggers
2026-03-31 21:28 ` Andrey Albershteyn [this message]
2026-04-01  6:28   ` [PATCH v6 06/22] iomap: introduce IOMAP_F_FSVERITY and teach writeback to handle fsverity Christoph Hellwig
2026-03-31 21:28 ` [PATCH v6 07/22] iomap: teach iomap to read files with fsverity Andrey Albershteyn
2026-03-31 23:30   ` Darrick J. Wong
2026-04-01  6:30   ` Christoph Hellwig
2026-03-31 21:28 ` [PATCH v6 08/22] iomap: introduce iomap_fsverity_write() for writing fsverity metadata Andrey Albershteyn
2026-03-31 23:32   ` Darrick J. Wong
2026-03-31 21:28 ` [PATCH v6 09/22] xfs: introduce fsverity on-disk changes Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 10/22] xfs: initialize fs-verity on file open Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 11/22] xfs: don't allow to enable DAX on fs-verity sealed inode Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 12/22] xfs: disable direct read path for fs-verity files Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 13/22] xfs: handle fsverity I/O in write/read path Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 14/22] xfs: use read ioend for fsverity data verification Andrey Albershteyn
2026-03-31 23:34   ` Darrick J. Wong
2026-03-31 21:28 ` [PATCH v6 15/22] xfs: add fs-verity support Andrey Albershteyn
2026-03-31 23:35   ` Darrick J. Wong
2026-04-01 23:57   ` Eric Biggers
2026-03-31 21:28 ` [PATCH v6 16/22] xfs: remove unwritten extents after preallocations in fsverity metadata Andrey Albershteyn
2026-03-31 23:36   ` Darrick J. Wong
2026-03-31 21:28 ` [PATCH v6 17/22] xfs: add fs-verity ioctls Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 18/22] xfs: advertise fs-verity being available on filesystem Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 19/22] xfs: check and repair the verity inode flag state Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 20/22] xfs: introduce health state for corrupted fsverity metadata Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 21/22] xfs: add fsverity traces Andrey Albershteyn
2026-04-01  6:31   ` Christoph Hellwig
2026-04-01 13:19     ` Andrey Albershteyn
2026-03-31 21:28 ` [PATCH v6 22/22] xfs: enable ro-compat fs-verity flag Andrey Albershteyn
2026-04-01  6:32   ` Christoph Hellwig
2026-04-01  6:32 ` [PATCH v6 00/22] fs-verity support for XFS with post EOF merkle tree Christoph Hellwig

Reply instructions:

You may reply publicly to this message via plain-text email
using any one of the following methods:

* Save the following mbox file, import it into your mail client,
  and reply-to-all from there: mbox

  Avoid top-posting and favor interleaved quoting:
  https://en.wikipedia.org/wiki/Posting_style#Interleaved_style

* Reply using the --to, --cc, and --in-reply-to
  switches of git-send-email(1):

  git send-email \
    --in-reply-to=20260331212827.2631020-7-aalbersh@kernel.org \
    --to=aalbersh@kernel.org \
    --cc=djwong@kernel.org \
    --cc=ebiggers@kernel.org \
    --cc=fsverity@lists.linux.dev \
    --cc=hch@lst.de \
    --cc=linux-btrfs@vger.kernel.org \
    --cc=linux-ext4@vger.kernel.org \
    --cc=linux-f2fs-devel@lists.sourceforge.net \
    --cc=linux-fsdevel@vger.kernel.org \
    --cc=linux-xfs@vger.kernel.org \
    /path/to/YOUR_REPLY

  https://kernel.org/pub/software/scm/git/docs/git-send-email.html

* If your mail client supports setting the In-Reply-To header
  via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox